/*
**  Sections
*/

class Section {
  constructor(headingElt) {
    /* headingElt is the element representing the section heading.
     * Can be a h[1-6], or a paragraph element.
     * Its next element sibling must be a div with the heading-body class.  */
    this.heading = headingElt.innerText;
    /* A list of elements that are owned directly by the section */
    this.contents = [];
    /* A list of child sections */
    this.childSections = [];
    /* Terminality - Whether the section can have child sections or not */
    this.isTerminal = headingElt.classList.contains("terminal-section");
    /* Depth - How many parent sections the section has */
    this.depth = calculate_heading_depth(headingElt);
    /* Whether the section has been rendered or not */
    this.rendered = false;
    /* Associate the heading element with its section */
    headingElt.associatedSection = this;
    this.headingElt = headingElt;
    /* Generate an ID for the section if it doesn't have one */
    if (headingElt.getAttribute("id") === null) {
      headingElt.setAttribute("id", generate_section_id(this.heading));
    }
    this.id = headingElt.getAttribute("id");
    /* Parse section body */
    const sectionBody = headingElt.nextElementSibling;
    parse_section_body(this, sectionBody);
  }
  generateNavPanelElements() {
    /* Returns a list item DOM element representing the section.
     * Used in navPanel construction. */
    let result = document.createElement("li");
    result.innerHTML = `<div>
                          ${this.childSections.length ?
                            "<button class=\"navpanel-toggle\"></button>" : ""}
                          <button class="section">${this.heading}</button>
                        </div>`;
    /* Associate the section with its button */
    let sectionButton = result.querySelector("button.section");
    sectionButton.section = this;
    this.button = sectionButton;
    /* Build elements for the child sections */
    if (this.childSections.length) {
      let sublist = document.createElement("ul");
      for (const childSection of this.childSections) {
        sublist.appendChild(childSection.generateNavPanelElements());
      }
      result.appendChild(sublist);
    }
    return result;
  }
  get_parents () {
    /* Get all parents of section up to and including topSection */
    let result = [];
    let section = this;
    while (section = section.parentSection) {
      result.push(section);
    }
    return result;
  }
  generate_link() {
    /* Generate a link representing given section */
    return `<a href="#${this.id}">${this.heading}</a>`;
  }
}

function element_is_or_has_heading(element) {
  let result = false;
  if (element.classList.contains("heading")) {
    /* Regular headings, h1-h6 */
    result = "is-heading";
  } else if ((element.nodeName === "OL") &&
             (element?.firstElementChild?.firstElementChild?.nodeName === "P") &&
             (element.firstElementChild.firstElementChild.classList.contains("heading"))) {
    /* Paragraph headings for heading levels beyond h6.
     * These are embedded inside ol, li elements. */
    result = "has-heading";
  }
  return result;
}

function calculate_heading_depth(element) {
  /* Calculates how many heading bodies the given element is nested in. */
  let result = 0;
  for (let p = element.parentElement; p; p = p.parentElement) {
    if (p.classList.contains("heading-body")) {
      result++;
    }
  }
  return result;
}

function parse_section_body(section, bodyElement) {
  /* Parses section body by building up section contents and child sections.
   * This function must live outside of the section class definition due to
   *   its use in the section constructor. */
  let subHeadingDetected = false;
  for (const childElt of bodyElement.children) {
    const childContainsSubHeading = element_is_or_has_heading(childElt);
    if (!childContainsSubHeading && !subHeadingDetected) {
      /* Parse section contents.
       * Content found after a subheading is detected is ignored, as this
       *   is not possible with org.
       * Elements are cloned while building contents, as the originals are
       *   from the content div, which must remain untouched to make document
       *   saving work. Reusing would cause rendering to remove them from
       *   their existing locations in the DOM. */
      section.contents.push(childElt.cloneNode(true));
    } else if (childContainsSubHeading) {
      /* A heading exists within childElt. */
      switch (childContainsSubHeading) {
        case "is-heading":
          add_heading_to_section(childElt, section);
          break;
        case "has-heading":
          for (const liElt of childElt.children) {
            if (liElt.firstElementChild?.classList.contains("heading")) {
              add_heading_to_section(liElt.firstElementChild, section);
            }
          }
          break;
        default:
          throw new Error();
      }
      subHeadingDetected = true;
    }
  }
}

function add_heading_to_section(headingElt, parentSection) {
  /*
   * Adds headingElt to parentSection either as a child section or directly
   *   in the parent's contents by checking for:
   * - whether parentSection is a terminal-section or not.
   * - whether headingElt is a non-section or not.
   * headingElt must be a H[1-6] or a P.
   */
  if ((headingElt.classList.contains("non-section"))
      || (parentSection.isTerminal)) {
    let promote = promote_heading(headingElt, parentSection.depth);
    parentSection.contents.push(...promote.promotedElements);
  } else {
    let newSection = new Section(headingElt)
    newSection.parentSection = parentSection;
    parentSection.childSections.push(newSection);
  }
}

function promote_heading(headingElt, promoteBy) {
  /*
   * Takes an element that is or has a heading (either H[1-6], P or an OL) and
   *   promotes the level of heading(s) associated with it by given promoteBy.
   * E.g. a h3 with promoteBy 2 yields a h1, a p promoted by 1 yields a h6
   *
   * Since an OL element can contain multiple headings, and since h[1-6] and p
   *   can have headings in their bodies that need to be promoted, this function
   *   returns an object containing the following information:
   * - elementsHandled :: Number of elements at headingElt's level that
   *   have been handled by the function.
   *   This is 2 for headers and paragraphs, and 1 for lists.
   * - promotedElements :: An array containing the promoted elements.
   *   This array contains clones, and does not reuse any of the original elts.
   */
  let result = {elementsHandled: 0, promotedElements: []};
  if (/H[0-9]/.test(headingElt.nodeName) || /P/.test(headingElt.nodeName)) {
    let newHeading = promote_header(headingElt, promoteBy);
    result.promotedElements.push(newHeading);
    let newBody = promote_heading_body(headingElt.nextElementSibling, promoteBy);
    result.promotedElements.push(newBody);
    result.elementsHandled = 2;
  } else if (headingElt.nodeName === "OL") {
    let currentLevel = calculate_heading_depth(headingElt) + 1;
    let newLevel = currentLevel - promoteBy;
    if (newLevel > 6) {
      /* No promotion needed - list stays a list */
      result.promotedElements.push(headingElt.cloneNode(true));
    } else {
      /* Promotion of paragraphs to header elements */
      for (const liElt of headingElt.children) {
        let pHeading = liElt.firstElementChild;
        let pBody = pHeading.nextElementSibling;
        result.promotedElements.push(promote_header(pHeading, promoteBy));
        result.promotedElements.push(promote_heading_body(pBody, promoteBy));
      }
    }
    result.elementsHandled = 1;
  } else {
    throw new Error();
  }
  return result;
}

function promote_header(headerElt, promoteBy) {
  /* headerElt can be a h[1-6] or a paragraph */
  let newHeading;
  let currentLevel = calculate_heading_depth(headerElt) + 1;
  let newLevel = currentLevel - promoteBy;
  if (newLevel < 1) {
    throw new Error();
  } else if (newLevel <= 6) {
    newHeading = document.createElement("H" + newLevel);
    newHeading.innerText = headerElt.innerText;
    clone_attributes(newHeading, headerElt);
  } else {
    newHeading = headerElt.cloneNode(true);
  }
  return newHeading;
}

function promote_heading_body(bodyElt, promoteBy) {
  let newBody = document.createElement("div");
  clone_attributes(newBody, bodyElt);
  let skip = 0;
  for (const childElt of bodyElt.children) {
    if (skip) {
      skip--;
      continue;
    }
    if (element_is_or_has_heading(childElt)) {
      let promote = promote_heading(childElt, promoteBy);
      promote.promotedElements.forEach((x) => newBody.appendChild(x));
      skip = promote.elementsHandled - 1;
    } else {
      newBody.appendChild(childElt.cloneNode(true));
    }
  }
  return newBody;
}

function clone_attributes(target, source) {
  [...source.attributes].forEach( attr => target.setAttribute(attr.nodeName, attr.nodeValue) )
}

function generate_section_id(heading_text) {
  /*
   * Generate a section id from given heading text.
   */
  /* Get the text up to the first opening paren */
  let result = heading_text.match(/[^\(]*/)[0];
  /* Replace spaces, remove non-alphanumerical characters */
  result = result.replace(/ /g, "-").replace(/[^0-9a-zA-Z\-]/g, "").toLowerCase();
  return result;
}

/*
**  NavPanel
*/

class NavPanel {
  constructor (navPanelElt) {
    /* Clear navPanel to get a clean slate if we're working with a saved file */
    navPanelElt.replaceChildren();
    /* Initialize */
    this.element = navPanelElt;
    this.element.appendChild(document.createElement("ul"));
  }
  addSection(section) {
    this.element.firstElementChild.appendChild(section.generateNavPanelElements());
  }
  finalize() {
    /* Set up navPanel scripts */
    let sublistInitialState;
    const sublists = this.element.querySelectorAll("ul ul");
    for (const sublist of sublists) {
      /* Make navPanel sublists smooth toggle targets */
      const toggleButton = sublist.previousElementSibling.firstElementChild;
      const sectionButton = toggleButton.nextElementSibling;
      sublistInitialState = "closed";
      if (sectionButton.section.headingElt.classList.contains("navpanel-open")) {
        sublistInitialState = "open";
      }
      st_setup(toggleButton, sublist, navpanel_button_get_target,
               navpanel_sublist_posttoggle, "vertical", sublistInitialState,
               undefined);
      st_setup(sectionButton, sublist, navpanel_button_get_target,
               navpanel_sublist_posttoggle, "vertical", sublistInitialState,
               undefined, true);
    }
    /* Set up section button scripts */
    const sectionButtons = this.element.querySelectorAll("button.section");
    for (const button of sectionButtons) {
      button.addEventListener("click", navpanel_section_button_click_handler);
    }
  }
}

function navpanel_button_get_target() {
  /* This needs to be a separate function to avoid the lexical scoping of the
   *   this keyword in arrow functions. */
  return this.parentElement.nextElementSibling;
}

function navpanel_sublist_posttoggle(toggle) {
  const button = this.previousElementSibling.firstElementChild;
  if (toggle === "open") {
    button.setAttribute("data-state", "open");
  } else {
    button.setAttribute("data-state", "closed");
  }
}

function navpanel_section_button_click_handler(event) {
  /* Render the section */
  if (this.st_getTarget &&
      (this.previousElementSibling.getAttribute("data-state") === "closed" ||
       mainRenderer.currentSection === this.section)) {
    st_smooth_toggle.call(this, event);
  }
  mainRenderer.render(this.section, null, "navpanel");
}

/*
**  MainRenderer
*/

class MainRenderer {
  constructor (mainElt) {
    /* Clear the main element to get a clean slate */
    mainElt.replaceChildren();
    /* Initialize */
    this.element = mainElt;
    this.header = document.createElement("h1");
    this.breadcrumbs = document.createElement("div");
    this.breadcrumbs.classList.add("main-breadcrumbs");
    this.bottom_navbar = document.createElement("div");
    this.bottom_navbar.classList.add("bottom-navbar");
  }
  render (section, id, source) {
    /*
     * - section :: Section to render.
     * - id      :: ID of the element to focus on. This can be the ID of an
     *              element within the section, or the ID of the section itself.
     * - source  :: Source of the render request, optional.
     *   Can be set to the following:
     *   - "renavigation" :: The render operation is a result of history
     *     navigation (back/forward buttons), thus should not manipulate
     *     the history stack.
     *   - "navpanel" :: The render operation is a result of navpanel navigation,
     *     thus should not touch navpanel state.
     *   - regexes :: The render operation is a result of search, thus the text
     *     matching the given regex should be highlighted.
     */
    /* Pre-render adjustments */
    if (!section.rendered) {
      section_prerender_process(section);
    }
    /* Render given section */
    this.render_breadcrumbs(section);
    this.header.innerText = section.heading;
    this.render_bottom_navbar(section);
    this.element.replaceChildren(this.breadcrumbs, this.header, ...section.contents, this.bottom_navbar);
    /* Post-render adjustments */
    if (!section.rendered) {
      section_postrender_process();
      section.rendered = true;
    }
    /* Deactivate the old section & activate the new section.
     * Need to account for topSection separately here, which does not have
     * a button but can be rendered. */
    if ((this.currentSection) && (this.currentSection.button)) {
      this.currentSection.button.parentElement.classList.remove("active");
    }
    this.currentSection = section;
    if (this.currentSection.button) {
      this.currentSection.button.parentElement.classList.add("active");
    }
    /* Highlight search terms */
    this.deactivate_highlight();
    if (source?.constructor.name === "RegExp") {
      this.activate_highlight(null, source);
      setTimeout(() => this.element.querySelector(".search-highlight")?.scrollIntoView({block: "center"}));
    }
    /* Prepare links */
    mainRenderer.element.querySelectorAll("a").forEach((x) => link_setup(x));
    /* If the sidebar is open, make sure that the button associated with the
     *   section is visible in the navpanel. */
    if (section.button && (source !== "navpanel") && !st_element_is_closed(sidebar)) {
      st_open_element(section.button.parentElement.parentElement.parentElement, true);
      section.button.parentElement.scrollIntoView();
    }
    /* Scrolling */
    let scrollTo = id ? this.element.querySelector(id) : null;
    if (scrollTo) {
      scrollTo.scrollIntoView({block: "center"});
    } else {
      window.scroll({top: 0, left: window.scrollX, behavior: "instant"});
    }
    /* Title and History Adjustment */
    if (!id) {
      id = "#" + section.id;
    }
    let idIsTopSection = (id === "#" + topSection.id);
    if ((source !== "renavigation") &&
        (!history.state || history.state.id !== id)) {
      history.pushState({id: id}, null,
                        (HOSTED ? host_hashed_link(id) :
                         (idIsTopSection ? window.location.pathname : id )));
    }
    document.title = ((idIsTopSection ? "" : section.heading + " - ") +
                      topSection.heading);
  }
  initialRender () {
    /*
     * Initial render:
     * - Render a section if the URL indicates it
     * - Render landing section if there is one
     * - Otherwise render topSection
     */
    let desiredSection = null;
    let source = "renavigation";
    if (HOSTED && INITIAL_RENDER) {
      desiredSection = find_section_from_id("#" + INITIAL_RENDER);
    } else if (window.location.hash) {
      desiredSection = find_section_from_id(window.location.hash);
    } else if (localStorage.getItem("visited") === null) {
      desiredSection = contentDiv.querySelector(".landing-section")?.associatedSection;
      if (desiredSection) {
        source = null;
      }
      localStorage.setItem("visited", "t");
    }
    mainRenderer.render(desiredSection ? desiredSection : topSection,
                        window.location.hash, source);
    /* Activate preferred theme */
    if (localStorage.getItem("theme")) {
      header.toggle_theme(localStorage.getItem("theme"), true)
    }
  }
  render_bottom_navbar (section) {
    /* Render the bottom navbar for given section according to its
     *   data-bottom-navbar property. */
    let navbar_spec = section.headingElt.getAttribute("data-bottom-navbar");
    let navbar_contents = null;
    if (navbar_spec) {
      let navbar_arr = navbar_spec.split(" ");
      if (navbar_arr.indexOf("none") !== -1) {
        ;
      } else if (navbar_arr.indexOf("children-direct") !== -1) {
        if (section.childSections.length) {
          let links = section.childSections.map((x) => `<li>${x.generate_link()}</li>`);
          navbar_contents = `<ol>${links.join("")}</ol>`
        }
      } else if (navbar_arr.indexOf("next") !== -1) {
        let next = find_next_section(section);
        if (next) {
          navbar_contents = `<span>Next:</span>${next.generate_link()}`;
        }
      } else {
        console.warn("Unknown data-bottom-navbar value: ", navbar_spec);
      }
    }
    if (navbar_contents) {
      this.bottom_navbar.innerHTML = navbar_contents;
      this.bottom_navbar.style.display = "block";
    } else {
      this.bottom_navbar.style.display = "none";
    }
  }
  render_breadcrumbs (section) {
    let parentLinks = [...section.get_parents().reverse(), section].map((x) => x.generate_link()).join(" » ");
    this.breadcrumbs.innerHTML = parentLinks;
    if (section != topSection) {
      this.breadcrumbs.style.display = "flex";
    } else {
      this.breadcrumbs.style.display = "none";
    }
  }
  activate_highlight (eltIterable, regex) {
    if (!eltIterable) {
      eltIterable = [this.header, ...this.currentSection.contents];
    }
    for (const elt of eltIterable) {
      if (elt.nodeType === Node.TEXT_NODE && regex.test(elt.textContent)) {
        let highlighted = elt.textContent.replace(regex, `<span class="search-highlight">$&</span>`);
        let newText = document.createElement("span");
        newText.classList.add("highlight-container");
        newText.innerHTML = highlighted;
        elt.replaceWith(newText);
      } else {
        this.activate_highlight(elt.childNodes, regex);
      }
    }
  }
  deactivate_highlight () {
    if (this.currentSection) {
      for (const elt of [this.header, ...this.currentSection.contents]) {
        for (const hcon of elt.querySelectorAll(".highlight-container")) {
          hcon.replaceWith(hcon.textContent);
        }
      }
    }
  }
}

function section_prerender_process(section) {
  /* Make layout adjustments required for collapsible-content */
  adjust_for_collapsible_content(section);
}

function adjust_for_collapsible_content(section, wrap) {
  /*
   * Recursively navigate the contents of given section to wrap
   *   the contents of any collapsible-content headings inside div's.
   */
  let contents_iterable = null;
  let wrapperDiv = null;
  let new_contents = [];
  if (section.constructor.name == "Section") {
    /* Section is a section */
    contents_iterable = section.contents;
  } else {
    /* Section is a DOM element */
    contents_iterable = Array.prototype.slice.call(section.children);
    if (wrap) {
      wrapperDiv = document.createElement("div");
      new_contents.push(wrapperDiv);
    }
  }
  let preheading = true;
  let heading_detected = false;
  for (const elt of contents_iterable) {
    heading_detected = element_is_or_has_heading(elt);
    if (preheading && wrap && !heading_detected) {
      wrapperDiv.appendChild(elt);
    } else {
      new_contents.push(elt);
      if (heading_detected) {
        preheading = false;
        let headings = get_headings(elt);
        for (const heading of headings) {
          adjust_for_collapsible_content(
            get_headings_body(elt, heading, contents_iterable),
            wrap || heading.classList.contains("collapsible-content"));
        }
      }
    }
  }
  if (wrap) {
    section.replaceChildren(...new_contents);
  }
}

function get_headings(element) {
  let result = [];
  if (element.classList.contains("heading")) {
    /* Regular headings, h1-h6 */
    result.push(element);
  } else if (element.nodeName === "OL") {
    /* Headings deeper than h6 */
    for (const liElt of element.children) {
      if (liElt?.firstElementChild.classList.contains("heading")) {
        result.push(liElt.firstElementChild);
      }
    }
  }
  return result;
}

function get_headings_body(elt, heading, contents_iterable) {
  let result = null;
  if (elt.classList.contains("heading")) {
    result = contents_iterable[contents_iterable.indexOf(elt) + 1];
  } else {
    result = heading.nextElementSibling;
  }
  return result;
}

function find_next_section(section) {
  /*
   * Find & return section next to the given one.
   * Next is determined according to the expanded navpanel order.
   * Can return null/undefined in case of no next section.
   */
  let result = null;
  if (section.childSections.length) {
    result = section.childSections[0];
  } else {
    /* Go up a level, get the sibling of section */
    let parent, children;
    while (section && !result) {
      parent = section.parentSection;
      children = parent.childSections;
      result = children[children.indexOf(section) + 1];
      if (!result) {
        section = parent;
      }
    }
  }
  return result;
}

function section_postrender_process() {
  /* Post-render adjustments for sections that have never been rendered. */
  /* Set up smooth-toggling for headings with the collapsible-body class */
  let colbody = mainRenderer.element.querySelectorAll(".collapsible-body")
  for (const heading of colbody) {
    st_setup(heading, heading.nextElementSibling, collapsible_body_get_target);
  }
  /* Set up smooth-toggling for collapsible-content headings */
  let colcont = mainRenderer.element.querySelectorAll(".collapsible-content, .collapsible-content + .heading-body .heading")
  for (const heading of colcont) {
    st_setup(heading, heading.nextElementSibling.firstElementChild, collapsible_content_get_target);
  }
}

function collapsible_body_get_target() {
  return this.nextElementSibling;
}

function collapsible_content_get_target() {
  return this.nextElementSibling.firstElementChild;
}

/*
**  Header
*/

class Header {
  constructor (headerElt) {
    this.element = headerElt;
    const landingButton = this.element.querySelector("button.manual-landing");
    landingButton.addEventListener("click", () => mainRenderer.render(topSection));
    this.sidebarToggle = this.element.querySelector("button.sidebar-toggle");
    this.themeToggle = this.element.querySelector("button.theme-toggle");

    /* Set up sidebar toggling */
    st_setup(this.sidebarToggle, sidebar, null, (t) => this.sidebar_posttoggle(t),
             "horizontal", "open", 320);

    /* Set up theme toggling */
    this.themeToggle.addEventListener("click", () => this.toggle_theme());
  }
  sidebar_posttoggle (toggle) {
    this.sidebarToggle.setAttribute("data-state", toggle === "open" ? "open" : "closed");
    /* If the sidebar is being opened, make sure that the current section
     *   is visible. */
    let section = mainRenderer.currentSection;
    if ((section?.button) && (toggle === "open")) {
      st_open_element(section.button.parentElement.parentElement.parentElement, true, false, sidebar);
      setTimeout(() => section.button.parentElement.scrollIntoView(), 100); /* 100 determined heuristically */
    }
    /* Reposition main if necessary */
    if (window.innerWidth >= WINDOW_WIDTH_THRESHOLD) {
      mainRenderer.element.style.transition = "left 0.2s";
      setTimeout(() => mainRenderer.element.style.transition = null, 200);
    }
    if (toggle === "open") {
      mainRenderer.element.style.left = "calc(var(--sidebar-width) + 2*var(--sidebar-offset-unit) + var(--main-offset-unit))";
    } else {
      mainRenderer.element.style.left = "calc((100vw - var(--main-width))/2)";
    }
  }
  toggle_theme (theme, immediate) {
    if (!theme) {
      theme = this.themeToggle.getAttribute("data-state") === "dark" ? "light" : "dark";
    }
    if (!immediate) {
      this.toggle_theme_transition_style("on");
      setTimeout(() => this.toggle_theme_transition_style("off"), 300);
    }
    this.themeToggle.setAttribute("data-state", theme);
    if (theme === "light") {
      document.querySelector("body").setAttribute("data-theme", theme);
    } else {
      document.querySelector("body").removeAttribute("data-theme");
    }
    localStorage.setItem("theme", theme);
  }
  toggle_theme_transition_style (toggle) {
    if (toggle === "on") {
      if (!document.querySelector("style.theme-transition-style")) {
        let ttStyle = document.createElement("style");
        ttStyle.classList.add("theme-transition-style");
        ttStyle.innerHTML = "* {transition: background-color 0.3s, border-color 0.3s !important; }"
        document.querySelector("head").appendChild(ttStyle);
      }
    } else if (toggle === "off") {
      document.querySelector("style.theme-transition-style")?.remove();
    }
  }
}

/*
**  Content
*/

function content_process(content_div) {
  /*
   * One-time adjustments to the content.
   */
  if (!content_div.classList.contains("processed")) {
    /* Avoid re-processing, e.g. in saved documents */
    content_div.classList.add("processed");

    /* Break down the collapsible-body-cascade class into its subclasses */
    let cbc_headings = content_div.querySelectorAll(".collapsible-body-cascade");
    for (const heading of cbc_headings) {
      heading.classList.add("non-section");
      heading.classList.add("collapsible-body");
      let subheadings = heading.nextElementSibling.querySelectorAll(".heading");
      for (const subheading of subheadings) {
        subheading.classList.add("collapsible-body");
      }
    }

    /* Break down the data-bottom-navbar="sequential" attribute */
    let sbn_headings = content_div.querySelectorAll('[data-bottom-navbar~="sequential"]');
    let apply_sequential = (heading) => {
      heading.setAttribute("data-bottom-navbar",
                           (heading.getAttribute("data-bottom-navbar") || "")
                           + " next");
    };
    for (const heading of sbn_headings) {
      let subheadings = heading.nextElementSibling.querySelectorAll(".heading");
      for (const subheading of subheadings) {
        apply_sequential(subheading);
      }
      apply_sequential(heading);
    }
  }
}

/*
**  Links
*/

function link_setup(link, callback) {
  /*
   * Link pre-processing.
   * - Translates internal links into hosted links if the document is hosted.
   * - Installs given callback as link click event handler.
   * Callback is optional, default is link_follow.
   */
  let href = link.getAttribute("href");
  if (HOSTED && /^#.*/.test(href)) {
    link.setAttribute("href", host_hashed_link(href));
  }
  if (!callback) {
    callback = link_follow;
  }
  link.addEventListener("click", callback);
}

function link_follow(event, matchRegex) {
  /*
   * matchRegex: Optional, regex to be highlighted on render if link is internal
   */
  let href = event.target.getAttribute("href");
  let id = (/^#.*/.test(href) ? href
            : (HOSTED && href.startsWith(HOSTURL.href) ?
               hash_hosted_link(href) : null));
  if (id) {
    /* Internal link */
    let section = find_section_from_id(id);
    if (section) {
      mainRenderer.render(section, id, matchRegex);
      event.preventDefault();
    }
  }
}

function find_section_from_id(id) {
  /*
   * Get the section that contains the element with given ID.
   * ID must begin with a hash.
   */
  let elt = contentDiv.querySelector(id);
  if (elt) {
    for (; elt != contentDiv; elt = elt.parentElement) {
      if (elt.associatedSection) {
        return elt.associatedSection;
      } else if (elt.classList.contains("heading-body") &&
                 elt.previousElementSibling.associatedSection) {
        return elt.previousElementSibling.associatedSection;
      }
    }
  }
}

/*
**  Window
*/

function window_setup() {
  /* Set up window events */
  window.addEventListener("popstate", rerender);

  function rerender(event) {
    let rendered = false;
    if (event.state) {
      let section = find_section_from_id(event.state.id);
      if (section) {
        mainRenderer.render(section, event.state.id, "renavigation");
        rendered = true;
      }
    }
    if (!rendered) {
      mainRenderer.render(topSection, null, "renavigation");
    }
  }
}

/*
**  Search Panel
*/

class SearchPanel {
  constructor (panelElt) {
    this.element = panelElt;
    this.searchInput = this.element.querySelector("input");
    this.quitButton = this.element.querySelector("button");
    this.resultsDiv = this.element.querySelector("div.results");
    let searchInputUpdate = () => {
      this.resultsDiv.replaceChildren();
      if (this.searchInput.value !== "") {
        let contextRegex = new RegExp(String.raw`[^\n]{0,20}${this.searchInput.value}[^\n]{0,20}`, "i");
        let matchRegex = new RegExp(this.searchInput.value, "ig");
        text_search(topSection, contextRegex, matchRegex);
        this.toggle(true, matchRegex);
      } else {
        this.toggle(false);
      }
    }
    this.searchInput.addEventListener("input", searchInputUpdate);
    this.searchInput.addEventListener("click", () => this.toggle(this.searchInput.value !== ""));
    this.quitButton.addEventListener("click", () => this.toggle(false));
    this.toggle(false);
  }
  add_search_result (section, contextStr, matchRegex) {
    let resultDiv = document.createElement("div");
    let searchBreadcrumbs = (section.get_parents().slice(0,-1).reverse()
                             .map((x) => x.generate_link()).join(" | "))
    contextStr = contextStr.replace(matchRegex, `<span class="search-highlight">$&</span>`);
    resultDiv.innerHTML = `${section.generate_link()}
                           ${searchBreadcrumbs ? '<div class="search-breadcrumbs">' + searchBreadcrumbs + "</div>" : "" }
                           <p>${contextStr}<p/>`
    let linkCallback = (e) => link_follow(e, matchRegex);
    resultDiv.querySelectorAll("a").forEach((x) => link_setup(x, linkCallback));
    this.resultsDiv.appendChild(resultDiv);
  }
  toggle (status, matchRegex) {
    if (status) {
      this.resultsDiv.style.display = "block";
      this.quitButton.style.display = "block";
      navPanel.element.style.visibility = "hidden";
      if (!matchRegex) {
        matchRegex = new RegExp(this.searchInput.value, "ig");
      }
      mainRenderer.deactivate_highlight();
      mainRenderer.activate_highlight(null, matchRegex);
    } else {
      this.resultsDiv.style.display = "none";
      this.quitButton.style.display = "none";
      navPanel.element.style.visibility = "visible";
      mainRenderer.deactivate_highlight();
    }
  }
}

function text_search(section, contextRegex, matchRegex) {
  let match = null;
  for (const content of [...section.contents, section.headingElt]) {
    match = content.innerText?.match(contextRegex);
    if (match) {
      searchPanel.add_search_result(section, match[0], matchRegex);
      break;
    }
  }
  for (const childSection of section.childSections) {
    text_search(childSection, contextRegex, matchRegex);
  }
}

/*
**  Hosted Documents
*/

function hosting_setup() {
  if (HOSTURL) {
    /* Add scheme if omitted */
    if (!/^https?:\/\//.test(HOSTURL)) { HOSTURL = "https://" + HOSTURL; }
    /* Remember if URL contains a trailing slash */
    window.HOSTURL_TRAILINGSLASH = /\/$/.test(HOSTURL);
    /* Translate HOSTURL from string to URL object */
    HOSTURL = new URL(HOSTURL);
  }
  /* Set up the HOSTED global */
  window.HOSTED = HOSTURL && (location.host === HOSTURL.host);
}

function host_hashed_link (id) {
  /*
   * If id belongs to a section, turns #id into HOSTURL.href/id
   * Otherwise turns #id into HOSTURL.href/section#id
   * Treats topSection as an edge case; returns HOSTURL.href or HOSTURL.href#id
   */
  let result, section;
  section = find_section_from_id(id);
  if (section === topSection) {
    result = HOSTURL.href + ((id !== ("#" + topSection.id)) ? id : "");
  } else {
    if (("#" + section.id) === id) {
      result = HOSTURL.href + (HOSTURL_TRAILINGSLASH ? "" : "/" ) + id.slice(1);
    } else {
      result = HOSTURL.href + (HOSTURL_TRAILINGSLASH ? "" : "/" ) + section.id + id;
    }
  }
  return result;
}

function hash_hosted_link (link) {
  /*
   * Turns HOSTURL.href/id and HOSTURL.href/section#id links into #id
   */
  let result, hashIndex;
  hashIndex = link.indexOf("#");
  if (hashIndex !== -1) {
    result = link.slice(hashIndex);
  } else {
    if (link === HOSTURL.href) {
      /* topSection */
      result = "#" + topSection.id;
    } else {
      result = "#" + link.slice(HOSTURL.href.length + !HOSTURL_TRAILINGSLASH);
    }
  }
  return result;
}

/*
**  Smooth Toggle Behavior
*/

function st_setup(st_button, st_target, findTargetFunc, postToggleFunc,
                  direction, initialState, openLength, customToggler) {
  /*
   * Set up smooth toggle behavior between a button and a target element.
   *
   * Arguments:
   * - st_button      :: Element that causes toggling.
   * - st_target      :: Element that gets expanded/collapsed.
   * - findTargetFunc :: Function that navigates from st_button to st_target
   *                     in the DOM tree. Optional, if unprovided a lambda is
   *                     generated.
   * - posttoggleFunc :: Function that handles post-toggle adjustments.
   *                     Called with context bound to the target element.
   *                     Optional.
   * - direction      :: Direction in which toggle happens. By default "vertical".
   *                     Can be set to "horizontal".
   * - initialState   :: Initial state of the target. Can be set to "open" or
   *                     "closed". Default is "closed".
   * - openLength     :: The length (width/height) of the target when it is open.
   *                     Must be an integer in pixel units.
   *                     If unset, the target is expanded fully.
   * - customToggler  :: If set to true, the default toggler function is not
   *                     installed as an event listener to the given button.
   *                     Indicates that a custom toggler will be installed.
   */
  /* Handle global st initialization */
  if (!Window.st_initializated) {
    st_global_initialize();
    Window.st_initializated = true;
  }

  /* Set up st for given button and target */
  st_target.classList.add("smooth-toggle-target");
  st_target.addEventListener("transitionend", st_transitionend);
  if (!customToggler) {
    st_button.addEventListener("click", st_smooth_toggle);
  }
  if (!findTargetFunc) {
    findTargetFunc = () => st_target;
  }
  st_button.st_getTarget = findTargetFunc;
  st_target.st_postToggle = postToggleFunc;
  st_target.st_direction = direction;
  if (direction === "horizontal") {
    st_target.classList.add("smooth-toggle-target-horizontal");
  } else {
    st_target.classList.add("smooth-toggle-target-vertical");
  }
  if (openLength) {
    st_target.st_maxLength = openLength;
  }
  if (initialState === "open") {
    st_open_element(st_target, true);
  } else {
    st_close_element(st_target, true);
  }
}

function st_global_initialize() {
  /* Called once on initialization for global adjustments required by st */
  /* Set up styles */
  if (!document.querySelector("style.smooth-toggle-style")) {
    let stStyle = document.createElement("style");
    stStyle.classList.add("smooth-toggle-style");
    stStyle.innerHTML = `.smooth-toggle-target-horizontal, .smooth-toggle-target-horizontal * {
                           overflow-x: hidden;
                           white-space: nowrap;
                         }
                         .smooth-toggle-target-vertical, .smooth-toggle-target-vertical * {
                           overflow-y: hidden;
                         }`
    document.querySelector("head").appendChild(stStyle);
  }
}

function st_getlength(elt) {
  /* Get the length of the element when it's open */
  return (elt.st_maxLength ? elt.st_maxLength :
          (elt.st_direction === "horizontal" ?
           elt.scrollWidth : elt.scrollHeight));
}

function st_calc_transition_duration(elt) {
  /* Calculate open/close transition duration of given element based on
   * its length */
  return Math.min(Math.max(st_getlength(elt) / 3000, 0.2), 1.0);
}

function st_element_is_closed(elt) {
  /* Checks if the element is closed or not.
   * maxHeight/maxWidth get set to an empty string if set to 0px from CSS,
   * and get set to 0px if transitioned from a nonzero value to 0px.
   * So we have to account for both. */
  return elt.st_direction === "horizontal" ?
    (!(elt.style.maxWidth)  || (elt.style.maxWidth === "0px")) :
    (!(elt.style.maxHeight) || (elt.style.maxHeight === "0px"));
}

function st_close_element(elt, now) {
  if (now) {
    elt.style.display = "none";
    if (elt.st_direction === "horizontal") {
      elt.style.maxWidth = "0";
    } else {
      elt.style.maxHeight = "0";
    }
    elt.transition = null;
  } else {
    let openLength = st_getlength(elt) + "px";
    if (elt.st_direction === "horizontal") {
      /* Set maxHeight to an explicit pixel value.
         Necessary because it can be set to max-content if the element's children
         have been expanded, which breaks transitions.*/
      elt.style.maxWidth = openLength;
      elt.style.transition = `max-width ${st_calc_transition_duration(elt)}s ease-out`;
      /* Initiate max width change in the next frame to make the transition work */
      setTimeout(() => elt.style.maxWidth = "0");
    } else {
      elt.style.maxHeight = openLength;
      elt.style.transition = `max-height ${st_calc_transition_duration(elt)}s ease-out`;
      setTimeout(() => elt.style.maxHeight = "0");
    }
  }
  if (elt.st_postToggle) {
    elt.st_postToggle("close");
  }
}

function st_open_element(elt, now, avoid_cascade, cascade_limit) {
  if (!avoid_cascade) {
    /* Cascade height change up the DOM tree */
    for (let p = elt.parentElement; p && p !== cascade_limit; p = p.parentElement) {
      if (p.classList.contains("smooth-toggle-target")) {
        st_open_element(p, true, true);
      }
    }
  }
  if (now) {
    elt.style.display = "block";
    let openLength = (elt.st_maxLength ? elt.st_maxLength + "px" : "max-content");
    elt.style.transition = null;
    if (elt.st_direction === "horizontal") {
      elt.style.maxWidth = openLength;
    } else {
      elt.style.maxHeight = openLength;
    }
  } else {
    elt.style.display = "block";
    let openLength = st_getlength(elt) + "px";
    if (elt.st_direction === "horizontal") {
      elt.style.transition = `max-width ${st_calc_transition_duration(elt)}s ease-out`;
      setTimeout(() => elt.style.maxWidth = openLength);
    } else {
      elt.style.transition = `max-height ${st_calc_transition_duration(elt)}s ease-out`;
      setTimeout(() => elt.style.maxHeight = openLength);
    }
  }
  if (elt.st_postToggle) {
    elt.st_postToggle("open");
  }
}

function st_transitionend() {
  /* Default transitionend handler for elements with smooth toggle behavior.
   * If the element is successfully closed, removes its contents from display.
   * Not doing so makes them remain visible to accessibility software and
   *   keyboard controls even if they're not rendered */
  if (st_element_is_closed(this)) {
    this.style.display = "none";
    /* /\* Close all descendants of the closed element as well *\/ */
    /* let descendants = this.querySelectorAll(".smooth-toggle-target"); */
    /* for (const elt of descendants) { */
    /*   st_close_element(elt, true); */
    /* } */
  }
}

function st_smooth_toggle(event) {
  /* Main click event handler for smooth toggle buttons */
  const toggleTarget = this.st_getTarget();
  if (toggleTarget.classList.contains("smooth-toggle-target")) {
    if (event.detail === 1) {
      /* Single clicks */
      if (!st_element_is_closed(toggleTarget)) {
        st_close_element(toggleTarget);
      } else {
        st_open_element(toggleTarget);
      }
    } else if (event.detail === 2) {
      /* Open heading body for the button and all its descendants */
      /* Open all descendant first to accurately calculate new length */
      let descendants = toggleTarget.querySelectorAll(".smooth-toggle-target");
      for (const elt of descendants) {
        st_open_element(elt, true, true);
      }
      st_open_element(toggleTarget);
    }
  }
}

/*
**  Imperative Code
*/

const WINDOW_WIDTH_THRESHOLD = 1184

hosting_setup();

const contentDiv = document.querySelector("div.content")
content_process(contentDiv);

const sidebar = document.querySelector("div.sidebar");
const navPanel = new NavPanel(document.querySelector("div.navpanel"));
const mainRenderer = new MainRenderer(document.querySelector("main"));
const header = new Header(document.querySelector("header"));
const searchPanel = new SearchPanel(document.querySelector("div.searchpanel"));

/* Build the top section, which will build the section tree of the manual */
const topSection = new Section(contentDiv.querySelector("h1"));
topSection.headingElt.setAttribute("data-bottom-navbar", "children-direct");

/* Build the navPanel using the immediate child sections of the topSection */
for (const childSection of topSection.childSections) {
  navPanel.addSection(childSection);
}
navPanel.finalize();

/* Set up window */
window_setup();

/* Initial render */
mainRenderer.initialRender();

/*
** Mobile Adjustments
*/

/* Start with sidebar closed on small screens */
if (window.innerWidth < WINDOW_WIDTH_THRESHOLD) {
  st_close_element(sidebar, true);
}

/* Clicking/pressing outside the sidebar should close the sidebar if it's open */
function avoid_sidebar_close() {
  if (window.innerWidth < WINDOW_WIDTH_THRESHOLD) {
    sidebar_close_avoided = true;
  }
}
sidebar.addEventListener("click", avoid_sidebar_close);
header.themeToggle.addEventListener("click", avoid_sidebar_close);
document.addEventListener("click", () => {
  if ((window.innerWidth < WINDOW_WIDTH_THRESHOLD)
      && !(st_element_is_closed(sidebar))
      && !sidebar_close_avoided) {
    st_close_element(sidebar);
  }
  sidebar_close_avoided = false;
});
